#!/bin/sh

: ${TMP:=/tmp}
: ${COMPAT_H:=fsc_compat.h}
: ${CC:=gcc}
: ${ARCH:=}
: ${HOSTCC:=}
: ${MACRO_PREFIX:=__FSC__}
EXTRA_CFLAGS="-I\$(TOPDIR) ${FSC_COMPAT_CFLAGS}"

ME=`basename $0`
# increment this whenever new tests are added or modified 
# (i.e. when the include files need to be modified)
MYVERSION=10039  # xx.yy.zz (no leading 0!!)

OLD=old_compat.h
TRUE=0
FALSE=1
   
myversion () {
    local mv
    mv=`echo $MYVERSION | sed 's/^\([0-9]\{1,2\}\)\([0-9]\{2\}\)\([0-9]\{2\}\)$/\1.\2.\3/'`
    [ -n "$mv" ] || die "invalid version"
    echo "$mv"
}          

basic_includes () {
    cat <<EOF
#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
EOF
}

include_if_avail () {
    has_header "$1" && echo "#include $1"
}

scsi_includes () {
    include_if_avail '<linux/blk.h>'
    include_if_avail '<linux/blkdev.h>'
    include_if_avail '<scsi/scsi.h>'
    include_if_avail '<scsi/scsi_host.h>'
    include_if_avail '<scsi/scsi_cmnd.h>'
    include_if_avail '<scsi/scsi_device.h>'
    include_if_avail '<scsi/scsi_dbg.h>'
    include_if_avail '"drivers/scsi/scsi.h"'
    include_if_avail '"drivers/scsi/sd.h"'
    include_if_avail '"drivers/scsi/hosts.h"'
}

diskdump_includes () {
    scsi_includes
    include_if_avail '<linux/diskdump.h>'
    include_if_avail '<linux/diskdump.h>'
    include_if_avail '<linux/diskdumplib.h>'
    include_if_avail '"drivers/scsi/scsi_dump.h"'
}

kobject_includes () {
    include_if_avail '<linux/kref.h>'
    include_if_avail '<linux/kobject.h>'
}

make_Makefile () {
   cat <<EOF
obj-m := `basename $TEST_BASE`.o
EXTRA_CFLAGS += ${EXTRA_CFLAGS}
ifeq (\${PATCHLEVEL},4)
include \$(TOPDIR)/Rules.make
endif
EOF
}

prepare_test () {
    TEST_BASE=$TEST_DIR/tst 
    rm -f $TEST_BASE.o $TEST_BASE.ko
    make_Makefile >$TEST_DIR/Makefile 
    basic_includes >$TEST_BASE.c
}

compile () {
    if [ -n "$ARCH" ]; then
	make -C $LX_DIR ARCH=$ARCH CC="$CC" HOSTCC="$HOSTCC" SUBDIRS=$TEST_DIR modules V=1 >/dev/null 2>&1
    else
	make -C $LX_DIR SUBDIRS=$TEST_DIR modules V=1 >/dev/null 2>&1
    fi
}


die () {
    echo "$ME: $@" >&2
    exit 129
}

text2macro () {
    # '"' -> 'q' (for header files)
    echo "$@" | tr 'a-z/.<>"-' A-Z____q_
}

set_macro () {
    if [ $2 -eq 0 ]; then
	eval "$1=1"
	echo "#define $1 1" >&3
	echo "yes"
    else
	eval "unset $1"
	echo "#undef  $1" >&3
	echo "no"
    fi    
}

do_check_noset () {
    prepare_test
    cat >>$TEST_BASE.c
    [ -n "$DEBUG" ] && cp -f $TEST_BASE.c $TEST_DIR/$1.c
    compile
    return $?
}

do_check () {
    local ret old
    old=`was_set "$1"`
    case $old in
	"") echo -n "checking for $1: "
	    do_check_noset $1
	    ret=$?
	    ;;
	NO) echo -n "caching $1: "; ret=1;;
	1)  echo -n "caching $1: "; ret=0;;
    esac
    set_macro "$1" $ret
    return $ret
}

force_check() {
    echo -n "forcing $1: "
    set_macro "$1" $2
}

header_macro_name () {
    echo "$MACRO_PREFIX"HEADER_`text2macro "$1"`
}

cpp_macro_name () {
    echo "$MACRO_PREFIX"CPP_`text2macro "$1"`
}

struct_macro_name () {
    echo "$MACRO_PREFIX"STRUCT_`text2macro "$1"`
}

enum_macro_name () {
    echo "$MACRO_PREFIX"ENUM_`text2macro "$1"`
}

type_macro_name () {
    echo "$MACRO_PREFIX"TYPE_`text2macro "$1"`
}

var_macro_name () {
    echo "$MACRO_PREFIX"VAR_`text2macro "$1"`
}

member_macro_name () {
    echo "$MACRO_PREFIX"`text2macro "$1"`_HAS_`text2macro "$2"`
}

no_func_macro_name () {
    echo "$MACRO_PREFIX"NO_FUNC_`text2macro "$1"`
}

check_header () {
    do_check `header_macro_name "$1"` <<EOF
$2
#include $1
EOF
}

# check if CPP macro is defined
check_cpp () {
    do_check `cpp_macro_name "$1"` <<EOF
$2
#ifndef $1
ERROR $1 undefined
#endif
EOF
}

check_enum () {
    do_check `enum_macro_name "$1"` <<EOF
$2
enum $1 _ts;
EOF
}

check_struct () {
    do_check `struct_macro_name "$1"` <<EOF
$2
struct $1 _ts;
EOF
}

check_type () {
    do_check `type_macro_name "$1"` <<EOF
$2
$1 _ts;
EOF
}

check_var () {
    do_check `var_macro_name "$1"` <<EOF
$2
void *__t = &$1;
EOF
}

check_scsi_struct () {
    check_struct "$1" "`scsi_includes`"
}

check_struct_member () {
    local ret name=`member_macro_name "$1" "$2"`
    case "`was_set $name`" in
    1)   echo -n "caching $name: "; set_macro $name 0; return;;
    NO)  echo -n "caching $name: "; set_macro $name 1; return;;
    esac
    do_check_noset $name <<EOF
$3
/* check for normal members */
void *_tf (struct $1 *s) { return &(s->$2); }
EOF
    ret=$?
    if [ $ret -eq 0 ]; then 
	echo -n "checking for $name: "
	set_macro $name $ret
    else
	do_check $name <<EOF
$3
/* check for bit fields if 1st test fails */
long _tf (struct $1 *s) { return (long)(s->$2); }
EOF
    fi
}

check_scsi_member () {
    check_struct_member "$1" "$2" "`scsi_includes`"
}

check_no_func () {
    do_check `no_func_macro_name "$1"` <<EOF
$2
/* This will generate an error if the function is defined */
struct __ts { char x[1]; };
char $1 (struct __ts *t) { return t->x[0];}
EOF
}

has () {
    eval "[ x\$$1 = x1 ]"
}

has_header () {
    has `header_macro_name "$1"`
}

has_struct () {
    has `struct_macro_name "$1"`
}

# If there was a fsc_compat.h file before, check old setting to skip test (speedup)
# returns empty string if not found, NO if #undef'd, value if #define'd
was_set () {
    local v=$1
    [ -f $TEST_DIR/$OLD ] || return
    sed -n "
/$1/ {
  s/^#define *$1 *\\([^ ]\\) *\$/\\1/p;
  s/^#undef *$1.*\$/NO/p;
  q;
}
"<$TEST_DIR/$OLD
}

uptodate () {
    local dir=$1
    [ -f $dir/$COMPAT_H ] && {
	cp -a $dir/$COMPAT_H $TEST_DIR/$OLD
	OLDVERSION=`sed -n 's/^#define __FSC_KERNEL_COMPAT_H *\([0-9]*\) *$/\1/p' <$dir/$COMPAT_H`
	[ -n "$OLDVERSION" ] && [ $OLDVERSION -ge $MYVERSION ]
    }
}

LX_DIR=$1
[ -n "$LX_DIR" ] || die "No kernel source directory specified"

TGT_DIR=$2
: ${TGT_DIR=$LX_DIR/include/linux}

[ -d $LX_DIR/include/linux ] || die "$LX_DIR: invalid directory"
[ -d $TGT_DIR ] || die "$TGT_DIR: invalid target directory"

if [ -n "$ARCH" ]; then
    [ -z "$CC" -o -z "$HOSTCC" ] && die "ARCH=$ARCH but CC/HOSTCC unset ($CC/$HOSTCC)"
fi

TEST_DIR=`mktemp -d $TMP/$ME.XXXXXX`
if [ -n "$DEBUG" ]; then
    echo "TEST_DIR=$TEST_DIR" >&2
else 
    trap "rm -rf $TEST_DIR"'*' 0
fi

[ "x$FORCE" = xyes ] && {
    echo FORCE is set, removing $TGT_DIR/$COMPAT_H >&2
    rm -f $TGT_DIR/$COMPAT_H 
}

uptodate $TGT_DIR  && {
#    echo "$ME: $TGT_DIR/$COMPAT_H is up-to-date (version `myversion`)."  >&2
    exit 0
}

echo "$ME: generating $TGT_DIR/$COMPAT_H (version `myversion`)" >&2
[ -n "$OLDVERSION" ] && echo "$ME: using previous version: $OLDVERSION" >&2
exec 3>$TEST_DIR/compat.h
check_header '<linux/sched.h>' || die "Cannot compile against headers in $LX_DIR"

cat >&3 <<EOF 
#ifndef __FSC_KERNEL_COMPAT_H
#define __FSC_KERNEL_COMPAT_H $MYVERSION
/* 
    FSC Kernel compatibility macros v`myversion`
    Autogenerated by $ME on `date +%Y-%m-%d` 
    for $LX_DIR

    These macros are generated from autoconf-like 
    kernel feature tests. Checking them is more reliable
    than using LINUX_VERSION_CODE or relying on the
    kernel programmers to supply feature check macros.

    Contact: <Martin.Wilck@Fujitsu-Siemens.com>
*/
EOF

check_header '<linux/device.h>' '' # 1.00.20
check_header '<linux/dma-mapping.h>' "`include_if_avail '<linux/device.h>'`"'
#include <linux/mm.h>
#include <asm/io.h>
#include <asm/scatterlist.h>
' # 1.00.19
check_header '<linux/kref.h>' "`kobject_includes`"
check_header '<linux/kobject.h>' "`kobject_includes`"
if has_header '<linux/kobject.h>'; then
    check_struct_member kobject kref "`kobject_includes`" 
    check_struct_member kobject refcount "`kobject_includes`"
fi

check_header '<linux/ioctl32.h>' # 1.00.19
check_header '<asm/ioctl32.h>'   # 1.00.19
check_header '<linux/moduleparam.h>'
check_type irqreturn_t '#include <linux/interrupt.h>
'; # 1.00.19

check_header '<linux/blk.h>'    # 1.00.11
check_header '<linux/blkdev.h>' # 1.00.11

check_header '<scsi/scsi.h>' "`scsi_includes`"
check_header '<scsi/scsi_host.h>' "`scsi_includes`"
check_header '<scsi/scsi_cmnd.h>' "`scsi_includes`"
check_header '<scsi/scsi_device.h>' "`scsi_includes`" # 1.0.14
check_header '<scsi/scsi_dbg.h>' "`scsi_includes`" # 1.0.18

# old scsi headers
check_header '"drivers/scsi/scsi.h"'  "`scsi_includes`" && \
    check_header '"drivers/scsi/sd.h"'  "`scsi_includes`"; \
    check_header '"drivers/scsi/hosts.h"'  "`scsi_includes`" && \
    check_header '"drivers/scsi/scsi_module.c"' "`scsi_includes`
extern Scsi_Host_Template driver_template;
"

check_scsi_struct Scsi_Host
check_scsi_struct SHT
check_scsi_struct scsi_host_template
check_scsi_struct scsi_cmnd # 1.0.28
check_type Scsi_Cmnd "`scsi_includes`"
check_type Scsi_Device "`scsi_includes`"           # 1.0.31
check_type Scsi_Host_Template "`scsi_includes`"    # 1.0.31
check_enum scsi_host_state "`scsi_includes`"    # 1.00.32

if has_struct SHT; then
    for f in highmem_io can_dma_32 single_sg_ok can_do_varyio vary_io need_plug_timer use_new_eh_code; do 
                                 # use_new_eh_code: 1.00.19
	check_scsi_member SHT $f
    done
fi
if has_struct scsi_host_template; then
    for f in dump_poll dump_sanity_check abort reset \
	eh_timed_out shost_attrs module; do  
# abort, reset: 1.00.12
# dump_sanity_check: 1.00.15
# eh_timed_out: 1.00.19
# shost_attrs: 1.00.19
# module: 1.00.19
	check_scsi_member scsi_host_template $f
    done
fi
if has_struct Scsi_Host; then
    check_scsi_member Scsi_Host lock
    check_scsi_member Scsi_Host host_lock
    check_scsi_member Scsi_Host can_queue_mask
    check_scsi_member Scsi_Host eh_active # 1.00.26
fi

if has_struct Scsi_Cmnd; then
    check_scsi_member scsi_cmnd state # 1.00.30
    check_scsi_member scsi_cmnd buffer # 1.00.37
fi

check_no_func scsi_sleep "`scsi_includes`"
check_no_func scsi_print_command "`scsi_includes`" # 1.00.18
check_no_func scsi_device_online "`scsi_includes`"  # 1.00.14
check_no_func scsi_set_device "`scsi_includes`"  # 1.00.19
check_no_func scsi_host_alloc "`scsi_includes`"  # 1.00.19
check_no_func scsi_scan_host "`scsi_includes`"  # 1.00.22

check_header '<linux/delay.h>' '' 
check_no_func msleep "`include_if_avail '<linux/delay.h>'`"
check_no_func ssleep "`include_if_avail '<linux/delay.h>'`"  # 1.00.34

check_header '<linux/jiffies.h>' '' 
check_no_func msecs_to_jiffies "`include_if_avail '<linux/jiffies.h>'`"'
#include <linux/time.h>
'

check_header '<linux/pm.h>' '' && { # 1.00.18
    check_type pm_message_t '#include <linux/pm.h> 
'
}
check_header '<linux/dmapool.h>' '' # 1.00.19
if check_header '<linux/pci.h>' ''   # 1.00.18
then
    check_struct_member pci_driver driver '#include <linux/pci.h> 
' # 1.00.19
    check_struct_member pci_driver shutdown '#include <linux/pci.h> 
' # 1.00.35
    check_no_func pci_disable_msi '#include <linux/pci.h> 
'
    check_no_func pci_set_consistent_dma_mask '#include <linux/pci.h> 
' # 1.00.23
else
    force_check $(no_func_macro_name pci_disable_msi) $TRUE
    force_check $(no_func_macro_name pci_set_consistent_dma_mask) $TRUE
fi

if check_header '<linux/diskdump.h>' "`diskdump_includes`" 
then
    check_header '<linux/diskdumplib.h>' "`diskdump_includes`"
    check_header '"drivers/scsi/scsi_dump.h"' "`diskdump_includes`" 
    check_no_func diskdump_sleep_on_timeout "`diskdump_includes`"
    check_no_func diskdump_ssleep "`diskdump_includes`" # 1.00.32
    check_no_func diskdump_mdelay "`diskdump_includes`" # 1.00.32
else
    force_check $(no_func_macro_name diskdump_sleep_on_timeout) $TRUE
    force_check $(no_func_macro_name diskdump_ssleep) $TRUE
    force_check $(no_func_macro_name diskdump_mdelay) $TRUE
fi

check_var netdump_mode "#include <linux/kernel.h>"  # 1.00.13
check_cpp __bitwise "#include <linux/types.h>"      # 1.00.16

if check_header "<linux/mii.h>"                     # 1.00.24
then
    check_no_func if_mii "#include <linux/mii.h>"   # 1.00.24
else
    force_check $(no_func_macro_name if_mii) $TRUE
fi

check_no_func find_task_by_pid "<linux/sched.h>" # 1.00.25
check_header "<asm/bootsetup.h>" # 1.00.32

if check_header "<linux/kthread.h>"               # 1.00.33
then
    check_no_func kthread_stop "#include <linux/kthread.h>"   # 1.00.33 
else
    force_check $(no_func_macro_name kthread_stop) $TRUE
fi

echo "#endif /* FSC_KERNEL_COMPAT_H */" >&3

mv $TEST_DIR/compat.h $TGT_DIR/$COMPAT_H
echo "$ME: done." >&2
